@@ -35,9 +35,15 @@ public:
   
   T operator()(int i) const
   {
+    if(i >= data.size()) throw std::runtime_error("Accessing index out of range matrix field");
     return *(data[i]);
   }
   
+  void sum(int i, const Eigen::MatrixXd& A)
+  {
+    *(data[i]) += A;
+  }
+  
   Eigen::RowVectorXd get_row(int n, int i) const{
     return data[n]->row(i);
   }
